/// Globally enable or disable [Ansi] settings
///
/// Handy for turning on and off embedded colors without commenting out code.
bool colorDisabled = false;

/// Pen attributes for foreground and background colors.
///
/// Use the pen in string interpolation to output ansi codes.
/// Use [up] in string interpolation to globally reset colors.
class Ansi {
  /// Treat a pen instance as a function such that pen("msg") is the same as
  /// pen.write("msg").
  String call(String msg) => write(msg);

  /// Allow pen colors to be used in a string.
  ///
  /// Note: Once the pen is down, its attributes remain in effect till they are
  /// changed by another pen or [up].
  @override
  String toString() {
    if (colorDisabled) return "";
    if (_pen != null) return _pen;

    final sb = new StringBuffer();
    if (_fcolor != null) {
      sb.write("${ansiEsc}38;5;${_fcolor}m");
    }

    if (_bcolor != null) {
      sb.write("${ansiEsc}48;5;${_bcolor}m");
    }

    return _pen = sb.toString();
  }

  /// Returns control codes to change the terminal colors.
  String get down => toString();

  /// Resets all pen attributes in the terminal.
  String get up => colorDisabled ? "" : ansiDefault;

  /// Write the [msg] with the pen's current settings and then reset all
  /// attributes.
  String write(String msg) => "${this}$msg$up";

  /// Set black color.
  void black({bool bg: false, bool bold: false}) => _std(0, bold, bg);

  /// Set red color.
  void red({bool bg: false, bool bold: false}) => _std(1, bold, bg);

  /// Set green color.
  void green({bool bg: false, bool bold: false}) => _std(2, bold, bg);

  /// Set yellow color.
  void yellow({bool bg: false, bool bold: false}) => _std(3, bold, bg);

  /// Set blue color.
  void blue({bool bg: false, bool bold: false}) => _std(4, bold, bg);

  /// Set magenta color.
  void magenta({bool bg: false, bool bold: false}) => _std(5, bold, bg);

  /// Set cyan color.
  void cyan({bool bg: false, bool bold: false}) => _std(6, bold, bg);

  /// Set white color.
  void white({bool bg: false, bool bold: false}) => _std(7, bold, bg);

  /// Sets the pen color to the rgb value between 0.0..1.0.
  void rgb({num r: 1.0, num g: 1.0, num b: 1.0, bool bg: false}) => xterm(
      (r.clamp(0.0, 1.0) * 5).toInt() * 36 +
          (g.clamp(0.0, 1.0) * 5).toInt() * 6 +
          (b.clamp(0.0, 1.0) * 5).toInt() +
          16,
      bg: bg);

  /// Sets the pen color to a grey scale value between 0.0 and 1.0.
  void gray({num level: 1.0, bool bg: false}) =>
      xterm(232 + (level.clamp(0.0, 1.0) * 23).round(), bg: bg);

  void _std(int color, bool bold, bool bg) =>
      xterm(color + (bold ? 8 : 0), bg: bg);

  /// Directly index the xterm 256 color palette.
  void xterm(int color, {bool bg: false}) {
    _pen = null;
    final c = color.toInt().clamp(0, 256);
    if (bg) {
      _bcolor = c;
    } else {
      _fcolor = c;
    }
  }

  ///Resets the pen's attributes.
  void reset() {
    _pen = null;
    _bcolor = _fcolor = null;
  }

  int _fcolor;
  int _bcolor;
  String _pen;
}

/// ANSI Control Sequence Introducer, signals the terminal for new settings.
String get ansiEsc => colorDisabled ? "" : '\x1B[';

/// Reset all colors and options for current SGRs to terminal defaults.
String get ansiDefault => colorDisabled ? "" : "${ansiEsc}0m";

/// Defaults the terminal's foreground color without altering the background.
/// Does not modify [Ansi]!
String resetForeground() => "${ansiEsc}39m";

/// Defaults the terminal's background color without altering the foreground.
/// Does not modify [Ansi]!
String resetBackground() => "${ansiEsc}49m";
